/*
 * Copyright (C) CNRS, INRIA, Université Bordeaux 1, Télécom SudParis
 * See COPYING in top-level directory.
 */

#include <stdio.h>
#include <GTG.h>
#include "eztrace_convert.h"

#include <starpu.h>
#include <starpu_task.h>

#include <core/jobs.h>

#include "starpu_ev_codes.h"
#include "eztrace_convert_starpu.h"


/* This structure contains information about the source of the starpu link */
struct starpu_link_source {
  char* source; /* The thread which submitted the task */
  varPrec time; /* date at which the task is submitted */
  struct starpu_task* task; /* Address of the task being submitted */
};

/* This list contains information about the source of the starpu link */
struct ezt_list_t starpu_link_source_list;

/* This structure contains the duration of a particular task */
struct starpu_task_duration {
  double start_time;
  struct starpu_task* task; /* Address of the task being executed */
};

/* This structure contains the durations of tasks of the same type */
struct starpu_task_type_stats {
  double min_time;
  double max_time;
  double ave_time;
  int nb_tasks;
  void* task_type;

  struct ezt_list_t starpu_task_duration_list; /* This list contains duration statistics for a particular task type */
};

/* This list contains duration statistics for a particular task type */
struct ezt_list_t starpu_task_type_stats_list;

char *jobName = NULL;

#define GET_PARAM_RAW_1(p_ev, arg1) do {		    \
    litl_read_get_param_1(p_ev, arg1);			    \
  } while(0)


static struct ezt_list_token_t* __find_matching_link_source(struct starpu_task * task)
{
  struct ezt_list_token_t *t = NULL;
  struct starpu_link_source *r;
  ezt_list_foreach(&starpu_link_source_list, t) {
    r = (struct starpu_link_source *) t->data;
    if(r->task == task)
    return t;
  }
  return t;
}

static struct ezt_list_token_t* __find_matching_task(app_ptr task, struct ezt_list_t *starpu_task_duration_list)
{
  struct ezt_list_token_t *t = NULL;
  struct starpu_task_duration *r;
  ezt_list_foreach(starpu_task_duration_list, t) {
    r = (struct starpu_task_duration *) t->data;
    if((app_ptr)r->task == task)
    return t;
  }
  return t;
}

static struct ezt_list_token_t* __find_matching_task_type(app_ptr task_type)
{
  struct ezt_list_token_t *t = NULL;
  struct starpu_task_type_stats *r;
  ezt_list_foreach(&starpu_task_type_stats_list, t) {
    r = (struct starpu_task_type_stats *) t->data;
    if((app_ptr)r->task_type == task_type)
    return t;
  }
  return t;
}


int eztrace_convert_starpu_init();
int handle_starpu_events(eztrace_event_t *ev);
void handleEZTRACE_starpu__starpu_driver_start_job() ;
void handleEZTRACE_starpu__starpu_driver_end_job() ;
void handleEZTRACE_starpu_starpu_task_submit_entry() ;
void handleEZTRACE_starpu_starpu_task_submit_exit() ;
void handleEZTRACE_starpu_starpu_task_wait_for_all_entry() ;
void handleEZTRACE_starpu_starpu_task_wait_for_all_exit() ;
void handleEZTRACE_starpu_callback_func_entry() ;
void handleEZTRACE_starpu_callback_func_exit() ;
void handleEZTRACE_starpu__starpu_deque_pop_task_10() ;


/*
 * This function will be called once all the plugins are loaded
 * and the trace is started.
 * This function usually declared StateTypes, LinkTypes, etc.
 */
int
eztrace_convert_starpu_init()
{
  if (get_mode() == EZTRACE_CONVERT) {
    addContType("CT_Scheduler", "CT_Program", "Scheduler");
    addContainer(0.00000, "CT_Sched", "CT_Scheduler", "C_Prog", "Scheduler", "0");
    addVarType("starpu_VAR__0", "submitted task", "CT_Process");
    addVarType("starpu_VAR__1", "pushed task", "CT_Process");
    addEntityValue("starpu_STATE_0", "ST_Thread", "Executing a task", GTG_GREEN);
    addEntityValue("starpu_STATE_1", "ST_Thread", "starpu_task_submit", GTG_TEAL);
    addEntityValue("starpu_STATE_2", "ST_Thread", "starpu_task_wait_for_all", GTG_ORANGE);
    addEntityValue("starpu_STATE_3", "ST_Thread", "callback_func", GTG_YELLOW);
    addEntityValue("starpu_STATE_4", "ST_Thread", "Sleeping", GTG_RED);
    addEntityValue("starpu_STATE_5", "ST_Thread", "Scheduling", GTG_BROWN);
    addEntityValue("starpu_STATE_6", "ST_Thread", "FetchingInput", GTG_BLUE);
    addLinkType ("L_StarPU_sub_exec", "StarPU submit-to-execute link", "CT_Program", "CT_Thread", "CT_Thread");
    addEventType("starpu_EVENT_0", "CT_Thread", "Steal task from another worker");
  }
  ezt_list_new(&starpu_link_source_list);
  ezt_list_new(&starpu_task_type_stats_list);
}

void handleEZTRACE_starpu_worker_set_status() {
    enum _starpu_worker_status status;

    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    GET_PARAM_PACKED_1(CUR_EV, status);
    if (status == STATUS_SLEEPING)
	CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_4");
    else if (status == STATUS_SCHEDULING)
	CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_5");
    else if (status == STATUS_UNKNOWN)
	CHANGE() popState (CURRENT, "ST_Thread", thread_id);
}

void  handleEZTRACE_starpu_worker_fetch_input_start() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_6");
}

void handleEZTRACE_starpu_worker_fetch_input_stop() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() popState (CURRENT, "ST_Thread", thread_id);
}


void handleEZTRACE_starpu_push_task() {
    static int flag = 0;

    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
      if (!flag) {
	flag = 1;
	CHANGE() setVar (CURRENT, "starpu_VAR__1", "CT_Sched", 0);
   }
      CHANGE() addVar (CURRENT, "starpu_VAR__1", "CT_Sched", (varPrec) 1);
}


int handle_starpu_stats(eztrace_event_t *ev) {
  recording_stats = 1;
  return handle_starpu_events(ev);
}

void print_starpu_stats() {
  printf("\nStarPU:\n");
  printf("-------\n");

  struct ezt_list_token_t *t = NULL;
  struct starpu_task_type_stats *r;
  ezt_list_foreach(&starpu_task_type_stats_list, t) {
    r = (struct starpu_task_type_stats *) t->data;
    r->ave_time /= (double)r->nb_tasks;
    printf("* Task type: %p\n", r->task_type);
    printf("   > Nb tasks = %d\n", r->nb_tasks);
    printf("   > Min time = %f ms\n", r->min_time);
    printf("   > Max time = %f ms\n", r->max_time);
    printf("   > Average time = %f ms\n", r->ave_time);

    ezt_list_remove(t);
    free(r);
    free(t);
  }

}

typedef struct s_nameList {
    char *name;
    struct s_nameList *next;
} nameList;

nameList *nameL = NULL;
nameList *savePtr;
nameList *nameExec = NULL;
nameList *execPtr;


void handleEZTRACE_starpu__starpu_driver_start_job() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    app_ptr task_ptr, func_ptr;
    nameList *tmp;
    char *link_id = NULL;
    struct starpu_link_source *source_data;
    char *execName;

    GET_PARAM_PACKED_2(CUR_EV, task_ptr, func_ptr);
//     CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_0");
    char *proc = calloc(1, sizeof(char) * 8);
    strcpy(proc, "CT_S");
    strncpy(proc+4, thread_id, 3);
    CHANGE() subVar (CURRENT, "starpu_VAR__0", proc, (varPrec) 1);
    if (strlen(jobName) > 1) {
	execName = malloc((strlen(jobName) + strlen("exec")) * sizeof(char));
	sprintf(execName, "exec%s", jobName);
	CHANGE() pushState (CURRENT, "ST_Thread", thread_id, execName);
    }
   else
       CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_0");
    struct starpu_task *t = (struct starpu_task *) task_ptr;
    struct ezt_list_token_t *token = __find_matching_link_source(t);

    if(token) {
	source_data = (struct starpu_link_source *) token->data;
	int ret = asprintf(&link_id, "%p", source_data->task);
	startLink (source_data->time, "L_StarPU_sub_exec", "C_Prog", source_data->source, thread_id, link_id, link_id);
	endLink (CURRENT, "L_StarPU_sub_exec", "C_Prog", source_data->source, thread_id, link_id, link_id);
//	if (strlen(jobName) > 1)
	//    free(execName);
    }

    /* Find the corresponding task list to update duration stats */
    struct ezt_list_token_t *fn_token = __find_matching_task_type(func_ptr);

    /* If the task type is new, create a new entry in the type list */
    if(!fn_token) {
	struct starpu_task_type_stats *new_type = malloc(sizeof(struct starpu_task_type_stats));
	new_type->min_time = 1000000.0;
	new_type->max_time = 0.0;
	new_type->ave_time = 0.0;
	new_type->nb_tasks = 0;
	new_type->task_type = (void*)func_ptr;
	ezt_list_new(&(new_type->starpu_task_duration_list));
	fn_token = malloc(sizeof(struct ezt_list_token_t));
	fn_token->data = new_type;
	ezt_list_add(&starpu_task_type_stats_list, fn_token);
    }

    struct starpu_task_type_stats *task_list;
    task_list = (struct starpu_task_type_stats *) fn_token->data;

    //* add the current task to the list
    struct starpu_task_duration *task_duration = malloc(sizeof(struct starpu_task_duration));
    task_duration->start_time = CURRENT;
    task_duration->task = (struct starpu_task *) task_ptr;

    struct ezt_list_token_t* to = malloc(sizeof(struct ezt_list_token_t));
    to->data = task_duration;
    ezt_list_add(&(task_list->starpu_task_duration_list), to);
    if (token) {
	ezt_list_remove(token);
	free(source_data);
	free(token);
//	free(link_id);
    }
}

void handleEZTRACE_starpu__starpu_driver_end_job() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    app_ptr task_ptr, func_ptr;

    GET_PARAM_PACKED_2(CUR_EV, task_ptr, func_ptr);
    CHANGE() popState (CURRENT, "ST_Thread", thread_id);

    /* Find the corresponding task list to update duration stats */
    struct ezt_list_token_t *fn_token = __find_matching_task_type(func_ptr);

    struct starpu_task_type_stats *task_list;
    task_list = (struct starpu_task_type_stats *) fn_token->data;

    /* Find the corresponding task in the list */
    struct ezt_list_token_t *token = __find_matching_task(task_ptr, &(task_list->starpu_task_duration_list));

    if(token) {
	struct starpu_task_duration *data;
	data = (struct starpu_task_duration *) token->data;
	double duration = CURRENT - data->start_time;
	if (duration < task_list->min_time)
	    task_list->min_time = duration;
	if (duration > task_list->max_time)
	    task_list->max_time = duration;
	task_list->ave_time += duration;
	(task_list->nb_tasks)++;
	ezt_list_remove(token);
	free(data);
	free(token);
    }

}

void handleEZTRACE_starpu_starpu_task_submit_entry() {
    nameList *tmp;
    char *printValue = malloc((strlen(jobName) + strlen("Task name: ")) * sizeof(char));
    char *execName;

    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
     if (nameL == NULL) {
	 nameL = malloc(sizeof(nameList));
	 savePtr = nameL;
	 nameL-> next = NULL;
	 nameL->name = strdup(jobName);
    } else {
	nameL->next = malloc(sizeof(nameList));
	nameL = nameL->next;
	nameL->name = strdup(jobName);
	nameL->next = NULL;
    }
    tmp = savePtr;
    while (tmp->next != NULL) {
	if (strcmp(tmp->name, jobName) == 0) {
	    CHANGE() pushState (CURRENT, "ST_Thread", thread_id, jobName);
	    return ;
	}
	tmp = tmp->next;
    }
    CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_1");
    sprintf(printValue, "Task name: %s", jobName);
    addEntityValue(jobName, "ST_Thread", printValue, GTG_PINK);
    printValue = malloc((strlen(jobName) + strlen("Executing task : ")) * sizeof(char));
    sprintf(printValue, "Executing task : %s", jobName);
    execName = malloc((strlen(jobName) + strlen("exec")) * sizeof(char));
    sprintf(execName, "exec%s", jobName);
    addEntityValue(execName, "ST_Thread", printValue, GTG_GREEN);
    free(printValue);
    free(execName);
}

void handleEZTRACE_starpu_starpu_task_submit_exit() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);

    app_ptr task_ptr;
    GET_PARAM_PACKED_1(CUR_EV, task_ptr);
    CHANGE() popState (CURRENT, "ST_Thread", thread_id);
    char *proc = calloc(1, sizeof(char) * 8);
    strcpy(proc, "CT_S");
    strncpy(proc+4, thread_id, 3);
    CHANGE() addVar (CURRENT, "starpu_VAR__0", proc, (varPrec) 1);

    //* add the current submission to the list

    struct starpu_link_source *source_data = malloc(sizeof(struct starpu_link_source));
    source_data->source = thread_id;
    source_data->time = CURRENT;
    source_data->task = (struct starpu_task *) task_ptr;

    struct ezt_list_token_t* token = malloc(sizeof(struct ezt_list_token_t));
    token->data = source_data;
    ezt_list_add(&starpu_link_source_list, token);
}

void handleEZTRACE_starpu_starpu_task_wait_for_all_entry() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_2");
}
void handleEZTRACE_starpu_starpu_task_wait_for_all_exit() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() popState (CURRENT, "ST_Thread", thread_id);
}
void handleEZTRACE_starpu_callback_func_entry() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() pushState (CURRENT, "ST_Thread", thread_id, "starpu_STATE_3");
}
void handleEZTRACE_starpu_callback_func_exit() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    CHANGE() popState (CURRENT, "ST_Thread", thread_id);
}
void handleEZTRACE_starpu__starpu_deque_pop_task_10() {
    FUNC_NAME;
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    int steal, workerid;
    app_ptr deque_queue;
    char* opt_value_EZTRACE_starpu__starpu_deque_pop_task_10 = NULL;

    GET_PARAM_PACKED_3(CUR_EV, deque_queue, workerid, steal);
    asprintf(&opt_value_EZTRACE_starpu__starpu_deque_pop_task_10, "deque_queue = %x, workerid = %x" , GET_PARAM(CUR_EV, 1), GET_PARAM(CUR_EV, 2) );
    if (steal) {
	CHANGE() addEvent (CURRENT, "starpu_EVENT_0", thread_id, opt_value_EZTRACE_starpu__starpu_deque_pop_task_10);
    }
    //free(opt_value_EZTRACE_starpu__starpu_deque_pop_task_10);
}

void handleStarpuInit() {
    char *proc = calloc(1, sizeof(char) * 4);
    char *procCpt = calloc(1, sizeof(char) * 8);
    DECLARE_THREAD_ID_STR(thread_id, CUR_INDEX, CUR_THREAD_ID);
    strcpy(procCpt, "CT_S");
    strncpy(proc, thread_id, 3);
    strcpy(procCpt+4, proc);
    addContainer(0.00000, procCpt, "CT_Scheduler", proc, "Task queue ", "0");
    CHANGE() setVar (CURRENT, "starpu_VAR__0", procCpt, 0);
}

/* This function is called by eztrace_convert for each event to
 * handle.
 * It shall return 1 if the event was handled successfully or
 * 0 otherwise.
 */
int
handle_starpu_events(eztrace_event_t *ev)
{
    switch (LITL_READ_GET_CODE(ev)) {
    case EZTRACE_starpu_init:
	handleStarpuInit();
	break;
    case EZTRACE_starpu__starpu_driver_start_job:
	handleEZTRACE_starpu__starpu_driver_start_job();
	break;
    case EZTRACE_starpu__starpu_driver_job_name:
	jobName = LITL_READ_RAW(CUR_EV)->data;
	break;
    case EZTRACE_starpu__starpu_driver_end_job:
	handleEZTRACE_starpu__starpu_driver_end_job();
	break;
    case EZTRACE_starpu_starpu_task_submit_name:
	jobName = LITL_READ_RAW(CUR_EV)->data;
	break;
    case EZTRACE_starpu_starpu_task_submit_entry:
	handleEZTRACE_starpu_starpu_task_submit_entry();
	break;
    case EZTRACE_starpu_starpu_task_submit_exit:
	handleEZTRACE_starpu_starpu_task_submit_exit();
	break;
    case EZTRACE_starpu_starpu_task_wait_for_all_entry:
	handleEZTRACE_starpu_starpu_task_wait_for_all_entry();
	break;
    case EZTRACE_starpu_starpu_task_wait_for_all_exit:
	handleEZTRACE_starpu_starpu_task_wait_for_all_exit();
	break;
    case EZTRACE_starpu_callback_func_entry:
	handleEZTRACE_starpu_callback_func_entry();
	break;
    case EZTRACE_starpu_callback_func_exit:
	handleEZTRACE_starpu_callback_func_exit();
	break;
    case EZTRACE_starpu__starpu_deque_pop_task_10:
	handleEZTRACE_starpu__starpu_deque_pop_task_10();
	break;
    case EZTRACE_starpu_push_task:
	handleEZTRACE_starpu_push_task();
	break;
    case EZTRACE_starpu_worker_set_status:
	handleEZTRACE_starpu_worker_set_status();
	break;
    case EZTRACE_starpu_worker_fetch_input_start:
	handleEZTRACE_starpu_worker_fetch_input_start();
	break;
    case EZTRACE_starpu_worker_fetch_input_stop:
	handleEZTRACE_starpu_worker_fetch_input_stop();
	break;
    default:
      /* The event was not handled */
	return 0;
    }
    return 1;
}



/* Constructor of the plugin.
 * This function registers the current module to eztrace_convert
 */
struct eztrace_convert_module starpu_module;
void libinit(void) __attribute__ ((constructor));
void libinit(void)
{
  starpu_module.api_version = EZTRACE_API_VERSION;

  /* Specify the initialization function.
   * This function will be called once all the plugins are loaded
   * and the trace is started.
   * This function usually declared StateTypes, LinkTypes, etc.
   */
  starpu_module.init = eztrace_convert_starpu_init;

  /* Specify the function to call for handling an event
   */
  starpu_module.handle = handle_starpu_events;

  starpu_module.handle_stats = handle_starpu_stats;
  starpu_module.print_stats = print_starpu_stats;

  /* Specify the module prefix */
  starpu_module.module_prefix = starpu_EVENTS_ID;

  asprintf(&starpu_module.name, "starpu");
  asprintf(&starpu_module.description, "Module for StarPU-based programs");

  starpu_module.token.data = &starpu_module;

  /* Register the module to eztrace_convert */
  eztrace_convert_register_module(&starpu_module);

  //printf("module starpu loaded\n");
}

void libfinalize(void) __attribute__ ((destructor));
void libfinalize(void)
{
  //printf("unloading module starpu\n");
}
