/*
 * Copyright (C) CNRS, INRIA, Université Bordeaux 1, Télécom SudParis
 * See COPYING in top-level directory.
 */

#include "eztrace.h"
#include "eztrace_sampling.h"

#include "starpu.h"
#include "starpu_task.h"
#include "core/jobs.h"

#include "starpu_ev_codes.h"

int filterMask = 0;

struct _starpu_work_stealing_data
{
	struct _starpu_deque_jobq **queue_array;
	unsigned rr_worker;
	/* keep track of the work performed from the beginning of the algorithm to make
	 * better decisions about which queue to select when stealing or deferring work
	 */
	unsigned performed_total;
	unsigned last_pop_worker;
	unsigned last_push_worker;
};

struct _starpu_job;
struct _starpu_worker;

int (*lib_starpu_push_task) (struct _starpu_job *j);

void (*lib_starpu_driver_start_job) (struct _starpu_worker* args, struct _starpu_job* j, struct timespec* codelet_start, int rank, int profiling);

void (*lib_starpu_driver_end_job) (struct _starpu_worker* args, struct _starpu_job* j, enum starpu_perfmodel_archtype perf_arch, struct timespec* codelet_end, int rank, int profiling);

int (*libstarpu_task_submit) (struct starpu_task* task);

int (*libstarpu_task_wait_for_all) ();

void (*libcallback_func) (void* callback_arg);

struct starpu_task * (*lib_starpu_deque_pop_task) (struct _starpu_deque_jobq* deque_queue, int workerid);

void (*lib_starpu_worker_set_status) (int workerid, enum _starpu_worker_status status);

int (*lib_starpu_fetch_task_input) (struct _starpu_job *j, uint32_t mask);

int (*libstarpu_init) (void *conf);

int _starpu_fetch_task_input(struct _starpu_job *j, uint32_t mask)
{
  int ret = 0;

  EZTRACE_EVENT_PACKED_0(EZTRACE_starpu_worker_fetch_input_start);
  ret = lib_starpu_fetch_task_input(j, mask);
  EZTRACE_EVENT_PACKED_0(EZTRACE_starpu_worker_fetch_input_stop);
  return ret;
}

void _starpu_worker_set_status (int workerid, enum _starpu_worker_status status)
{
  EZTRACE_EVENT_PACKED_1(EZTRACE_starpu_worker_set_status, status);
  lib_starpu_worker_set_status(workerid, status);
  }

int _starpu_push_task(struct _starpu_job *j) {
  FUNCTION_ENTRY;
  EZTRACE_EVENT_PACKED_0(EZTRACE_starpu_push_task);
  int ret = lib_starpu_push_task(j);
  return ret;
}

void _starpu_driver_start_job (struct _starpu_worker* args, struct _starpu_job* j, struct timespec* codelet_start, int rank, int profiling) {
    char name[] = "StarPU internal";
    FUNCTION_ENTRY;

    if (j->task->cl->name != NULL) {
	EZTRACE_EVENT_RAW (EZTRACE_starpu__starpu_driver_job_name, (char*)j->task->cl->name);
    } else {
	EZTRACE_EVENT_RAW (EZTRACE_starpu__starpu_driver_job_name, name);
	}
    EZTRACE_EVENT_PACKED_2 (EZTRACE_starpu__starpu_driver_start_job, j->task, j->task->cl->cpu_funcs[0]);
    lib_starpu_driver_start_job (args, j, codelet_start, rank, profiling);
}

void _starpu_driver_end_job (struct _starpu_worker* args, struct _starpu_job* j, enum starpu_perfmodel_archtype perf_arch, struct timespec* codelet_end, int rank, int profiling) {
    FUNCTION_ENTRY;
    EZTRACE_EVENT_PACKED_2 (EZTRACE_starpu__starpu_driver_end_job, j->task, j->task->cl->cpu_funcs[0]);
    lib_starpu_driver_end_job (args, j, perf_arch, codelet_end, rank, profiling);
}

int starpu_task_submit (struct starpu_task* task) {
  FUNCTION_ENTRY
     char name[] = "StarPU internal";
   if (task->cl != NULL && task->cl->name != NULL) {
       EZTRACE_EVENT_RAW (EZTRACE_starpu_starpu_task_submit_name, (char*)task->cl->name);
    } else {
	EZTRACE_EVENT_RAW (EZTRACE_starpu_starpu_task_submit_name, name);
	}
    EZTRACE_EVENT_PACKED_1 (EZTRACE_starpu_starpu_task_submit_entry, task);
    int ret = libstarpu_task_submit (task);
    EZTRACE_EVENT_PACKED_1 (EZTRACE_starpu_starpu_task_submit_exit, task);
    return ret;
}

int starpu_task_wait_for_all () {
  FUNCTION_ENTRY
	EZTRACE_EVENT_PACKED_0 (EZTRACE_starpu_starpu_task_wait_for_all_entry);
	int ret = libstarpu_task_wait_for_all ();
	EZTRACE_EVENT_PACKED_0 (EZTRACE_starpu_starpu_task_wait_for_all_exit);
	return ret;
}

void callback_func (void* callback_arg) {
  FUNCTION_ENTRY
	EZTRACE_EVENT_PACKED_1 (EZTRACE_starpu_callback_func_entry, callback_arg);
	libcallback_func (callback_arg);
	EZTRACE_EVENT_PACKED_1 (EZTRACE_starpu_callback_func_exit, callback_arg);
}


struct starpu_task * _starpu_deque_pop_task (struct _starpu_deque_jobq* deque_queue, int workerid) {
  FUNCTION_ENTRY
	struct _starpu_work_stealing_data *ws = (struct _starpu_work_stealing_data*)starpu_sched_ctx_get_policy_data(0);
  app_ptr q = (app_ptr)ws->queue_array[workerid];

	struct starpu_task * ret = lib_starpu_deque_pop_task (deque_queue, workerid);
	int steal = ((struct _starpu_deque_jobq*)q != deque_queue && ret != NULL);
	EZTRACE_EVENT_PACKED_3 (EZTRACE_starpu__starpu_deque_pop_task_10, deque_queue, workerid, steal);
	return ret;
}

int starpu_init(struct starpu_conf *conf) {
    FUNCTION_ENTRY
	EZTRACE_EVENT_PACKED_0 (EZTRACE_starpu_init);
    int ret = libstarpu_init(conf);
    return ret;
}


START_INTERCEPT
INTERCEPT2("starpu_init", libstarpu_init)

INTERCEPT2("_starpu_worker_set_status", lib_starpu_worker_set_status)

    INTERCEPT2("_starpu_driver_start_job", lib_starpu_driver_start_job)

    INTERCEPT2("_starpu_driver_end_job", lib_starpu_driver_end_job)

    INTERCEPT2("_starpu_fetch_task_input", lib_starpu_fetch_task_input)

    INTERCEPT2("_starpu_push_task", lib_starpu_push_task)

    INTERCEPT2("starpu_task_submit", libstarpu_task_submit)

    INTERCEPT2("_starpu_deque_pop_task", lib_starpu_deque_pop_task)

    INTERCEPT2("starpu_task_wait_for_all", libstarpu_task_wait_for_all)

    INTERCEPT2("callback_func", libcallback_func)
END_INTERCEPT

//INTERCEPT2("_starpu_worker_restart_sleeping", lib_starpu_worker_restart_sleeping)
// INTERCEPT2("_starpu_worker_stop_sleeping", lib_starpu_worker_stop_sleeping)

static void __starpu_init (void) __attribute__ ((constructor));
/* Initialize the current library */
static void
__starpu_init (void)
{
  DYNAMIC_INTERCEPT_ALL();

  /* start event recording */
#ifdef EZTRACE_AUTOSTART
  eztrace_start ();
#endif



}

static void __starpu_conclude (void) __attribute__ ((destructor));
static void
__starpu_conclude (void)
{
  /* stop event recording */
  eztrace_stop ();
}
